structure Semant = 

struct

structure A = AbstractSyntax;

structure AB = AssumptionBase;

datatype value = propVal of A.prop
               | primMethodVal of A.prim_method
               | primFunVal of A.prim_fun
               | funClosVal of {params:A.ide list,body:A.exp,env:environment ref} 
               | methClosVal of {params:A.ide list,body:A.ded,env:environment ref}
and 
   env_binding = unBound | bind of value

withtype environment = A.ide -> env_binding;

fun printVal(propVal(P)) = A.printProp(P)
  | printVal(primMethodVal(m)) = A.printPrimMethod(m)
  | printVal(primFunVal(f)) = A.printPrimFun(f)
  | printVal(funClosVal(_)) = "Function"
  | printVal(methClosVal(_)) = "Method"

exception EvalError of string;

fun evError(str) = raise EvalError("\n"^str^"\n");

val empty_env = fn id => unBound;

fun extend(f,[]) = f
  | extend(f,(a,b)::rest) = extend(fn x => if x = a then b else f(x),rest);

fun augmentEnv(env1,env2) = fn id => (case env2(id) of
                                         unBound => env1(id)
                                       | b as bind(v) => b);
fun failMatch() = evError("Failed match.");

fun matchPropToPat(P,A.idPat(id),env) = 
       (case env(id) of
           unBound => extend(env,[(id,bind(propVal(P)))])
         | bind(propVal(Q)) => if P = Q then env else failMatch()
         | _ => failMatch())
  | matchPropToPat(_,A.anyPat,env) = env
  | matchPropToPat(A.trueProp,A.truePat,env) = env
  | matchPropToPat(A.falseProp,A.falsePat,env) = env
  | matchPropToPat(A.atom(str),A.atomPat(str'),env) = if str = str' then env else failMatch()
  | matchPropToPat(A.neg(P),A.negPat(pat),env) = matchPropToPat(P,pat,env)
  | matchPropToPat(A.conj(P1,P2),A.conjPat(pat1,pat2),env) = 
      matchPropToPat(P2,pat2,matchPropToPat(P1,pat1,env))
  | matchPropToPat(A.disj(P1,P2),A.disjPat(pat1,pat2),env) = 
      matchPropToPat(P2,pat2,matchPropToPat(P1,pat1,env))
  | matchPropToPat(A.cond(P1,P2),A.condPat(pat1,pat2),env) = 
      matchPropToPat(P2,pat2,matchPropToPat(P1,pat1,env))
  | matchPropToPat(A.biCond(P1,P2),A.biCondPat(pat1,pat2),env) = 
      matchPropToPat(P2,pat2,matchPropToPat(P1,pat1,env))
  | matchPropToPat(_) = failMatch();


fun matchPropToPatWCost(P,A.idPat(id),env,c) = 
       (case env(id) of
           unBound => (SOME(extend(env,[(id,bind(propVal(P)))])),c+1)
         | bind(propVal(Q)) => if P = Q then (SOME(env),c+2) else (NONE,c+2)
         | _ => (NONE,c+1))
  | matchPropToPatWCost(_,A.anyPat,env,c) = (SOME(env),c+1)
  | matchPropToPatWCost(A.trueProp,A.truePat,env,c) = (SOME(env),c+1)
  | matchPropToPatWCost(A.falseProp,A.falsePat,env,c) = (SOME(env),c+1)
  | matchPropToPatWCost(A.atom(str),A.atomPat(str'),env,c) = 
         if str = str' then (SOME(env),c+2) else (NONE,c+2)
  | matchPropToPatWCost(A.neg(P),A.negPat(pat),env,c) = matchPropToPatWCost(P,pat,env,c+1)
  | matchPropToPatWCost(A.conj(P1,P2),A.conjPat(pat1,pat2),env,c) = 
       (case matchPropToPatWCost(P1,pat1,env,c) of
           (SOME(env'),c') => matchPropToPatWCost(P2,pat2,env',c')
         | (_,c') => (NONE,c'))
  | matchPropToPatWCost(A.disj(P1,P2),A.disjPat(pat1,pat2),env,c) = 
       (case matchPropToPatWCost(P1,pat1,env,c) of
           (SOME(env'),c') => matchPropToPatWCost(P2,pat2,env',c')
         | (_,c') => (NONE,c'))
  | matchPropToPatWCost(A.cond(P1,P2),A.condPat(pat1,pat2),env,c) = 
       (case matchPropToPatWCost(P1,pat1,env,c) of
           (SOME(env'),c') => matchPropToPatWCost(P2,pat2,env',c')
         | (_,c') => (NONE,c'))
  | matchPropToPatWCost(A.biCond(P1,P2),A.biCondPat(pat1,pat2),env,c) = 
       (case matchPropToPatWCost(P1,pat1,env,c) of
           (SOME(env'),c') => matchPropToPatWCost(P2,pat2,env',c')
         | (_,c') => (NONE,c'))
  | matchPropToPatWCost(_,_,_,c) = (NONE,c);

fun findMatch(P,cases) =
      let fun tryCases([]) = failMatch()
            | tryCases((pat,phr)::rest) = 
                (matchPropToPat(P,pat,empty_env),phr) handle _ => tryCases(rest)
      in
         tryCases(cases)
      end;

fun findMatchAndCost(P,cases) =
      let fun tryCases([],c) = failMatch()
            | tryCases((pat,phr)::rest,c) = 
                (case matchPropToPatWCost(P,pat,empty_env,c) of
                    (SOME(env'),c') => (env',phr,c')
                  | (_,c') => tryCases(rest,c'))
      in
         tryCases(cases,0)
      end;

fun checkProps(props,ab,name) = 
      let fun f([]) = ()
            | f(P::rest) = if AB.member(P,ab) then f(rest) else 
                           evError("Invalid application of "^name^": the proposition "^
                                   A.printProp(P)^" is not in the assumption base.")
      in
         f(props)
      end;
 

fun applyPrimMethod(A.claim,[v as propVal(P)],ab) = (checkProps([P],ab,"claim");v)
  | applyPrimMethod(A.dn,[propVal(P as A.neg(A.neg(Q)))],ab) = 
        (checkProps([P],ab,"dn");propVal(Q))
  | applyPrimMethod(A.mp,[propVal(P1 as A.cond(P,Q)),propVal(P2)],ab) =  
        if P2 = P then (checkProps([P1,P2],ab,"mp");propVal(Q))
        else evError("Ill-formed application of mp.")
  | applyPrimMethod(A.both,[propVal(P1),propVal(P2)],ab) = 
        (checkProps([P1,P2],ab,"both");propVal(A.conj(P1,P2)))
  | applyPrimMethod(A.leftAnd,[propVal(P as A.conj(P1,_))],ab) =  
        (checkProps([P],ab,"left-and");propVal(P1))
  | applyPrimMethod(A.leftEither,[propVal(P),propVal(Q)],ab) = 
        (checkProps([P],ab,"left-either");propVal(A.disj(P,Q)))
  | applyPrimMethod(A.rightEither,[propVal(P),propVal(Q)],ab) = 
        (checkProps([Q],ab,"right-either");propVal(A.disj(P,Q)))
  | applyPrimMethod(A.rightAnd,[propVal(P as A.conj(_,P2))],ab) = 
        (checkProps([P],ab,"right-and");propVal(P2))
  | applyPrimMethod(A.cd,[propVal(P as A.disj(P1,P2)),
                         propVal(Q as A.cond(P3,P4)),propVal(R as A.cond(P5,P6))],ab) =
       if P3 = P1 andalso P5 = P2 andalso P4 = P6 then
          (checkProps([P,Q,R],ab,"cd");propVal(P4))
       else evError("Ill-formed application of cd")
  | applyPrimMethod(A.equiv,[propVal(P as A.cond(P1,P2)),propVal(Q as A.cond(P3,P4))],ab) = 
      if P3 = P2 andalso P4 = P1 then
         (checkProps([P,Q],ab,"equiv");propVal(A.biCond(P1,P2)))
      else evError("Ill-formed application of equiv")
  | applyPrimMethod(A.leftIff,[propVal(P as A.biCond(P1,P2))],ab) = 
           (checkProps([P],ab,"left-iff");propVal(A.cond(P1,P2)))
  | applyPrimMethod(A.rightIff,[propVal(P as A.biCond(P1,P2))],ab) = 
           (checkProps([P],ab,"right-iff");propVal(A.cond(P2,P1)))
  | applyPrimMethod(A.absurd,[propVal(P as P1),propVal(Q as A.neg(P2))],ab) = 
      if P2 = P1 then (checkProps([P,Q],ab,"absurd");propVal(A.falseProp))
      else evError("Ill-formed application of absurd")
  | applyPrimMethod(A.trueIntro,[],_) = propVal(A.trueProp)
  | applyPrimMethod(A.negRule,[propVal(P)],ab) = 
            (checkProps([A.falseProp],ab,"neg");propVal(A.neg(P)))
  | applyPrimMethod(A.condRule,[propVal(P),propVal(Q)],ab) = 
            (checkProps([Q],ab,"cond");propVal(A.cond(P,Q)))
  | applyPrimMethod(_) = evError("Ill-formed primitive method application.");
      

fun applyPrimFun(A.eqFun,[propVal(P),propVal(Q)],_) = 
       if P = Q then propVal(A.trueProp) else propVal(A.falseProp)
  | applyPrimFun(A.notFun,[propVal(P)],_) = propVal(A.neg(P))
  | applyPrimFun(A.andFun,[propVal(P),propVal(Q)],_) = propVal(A.conj(P,Q)) 
  | applyPrimFun(A.orFun,[propVal(P),propVal(Q)],_) = propVal(A.disj(P,Q)) 
  | applyPrimFun(A.ifFun,[propVal(P),propVal(Q)],_) = propVal(A.cond(P,Q)) 
  | applyPrimFun(A.iffFun,[propVal(P),propVal(Q)],_) = propVal(A.biCond(P,Q))
  | applyPrimFun(f,_,_) = evError("Invalid application of "^A.printPrimFun(f));

fun evExp(A.constExp(A.primMethodConst(m)),_,_) = primMethodVal(m)
      | evExp(A.constExp(A.primFunConst(f)),_,_) = primFunVal(f)
      | evExp(A.constExp(A.trueConst),_,_) = propVal(A.trueProp)
      | evExp(A.constExp(A.falseConst),_,_) = propVal(A.falseProp)
      | evExp(A.constExp(A.propAtomConst(A)),_,_) = propVal(A.atom(A))
      | evExp(A.idExp(id),env,_) = (case (!env)(id) of 
                                       unBound => evError("Unbound identifier: "^id)
                                     | bind(v) => v)
      | evExp(A.funExp({params,body}),cenv,_) = funClosVal({params=params,body=body,env=cenv})
      | evExp(A.methodExp({params,body}),cenv,_) = methClosVal({params=params,body=body,env=cenv})
      | evExp(A.funAppExp(fexp,args),env,ab) = 
            let val (fval,arg_vals) = (evExp(fexp,env,ab),map (fn p => evPhrase(p,env,ab)) args)
            in
              (case fval of
                  primFunVal(f) => applyPrimFun(f,arg_vals,ab)
                | funClosVal({params,body,env=cenv}) => 
                    evExp(body,ref(extend(!cenv,ListPair.zip(params,map bind arg_vals))),ab)
                | _ => evError("Non-function value in the leftmost position of a function application."))
            end
      | evExp(A.letExp([],body),env,ab) = evExp(body,env,ab)
      | evExp(A.letExp((id,phr)::rest,body),env,ab) = 
            evExp(A.letExp(rest,body),ref(extend(!env,[(id,bind(evPhrase(phr,env,ab)))])),ab)
      | evExp(A.fixExp(x,body),env,ab) =
          let val rec_env = ref(!env)
              val rec_value = evExp(body,rec_env,ab)
              val new_env = extend(!env,[(x,bind rec_value)])
              val _ = rec_env := new_env
          in
             rec_value
          end
      | evExp(A.matchExp({discriminant,cases}),env,ab) = 
          (case evExp(discriminant,env,ab) of 
               propVal(P) => 
                 let val (env',e) = findMatch(P,map (fn (pat,e) => (pat,A.expression(e))) cases)
                 in
                   evPhrase(e,ref(augmentEnv(!env,env')),ab)
                 end
            | _ => evError("Attempt to pattern-match a value that is not a proposition."))
      | evExp(A.seqExp([]),_,_) = evError("Empty expression sequence.")
      | evExp(A.seqExp([e]),env,ab) = evExp(e,env,ab)
      | evExp(A.seqExp(e::rest),env,ab) = (evExp(e,env,ab);evExp(A.seqExp(rest),env,ab))
    and
        evDed(A.methodAppDed(mexp,args),env,ab) =  
           let val mval = evExp(mexp,env,ab)
               fun getArgVals([],lemmas,results) = (rev(results),lemmas)
                 | getArgVals(A.expression(e)::rest,lemmas,results) = 
                       getArgVals(rest,lemmas,evExp(e,env,ab)::results)
                 | getArgVals(A.deduction(d)::rest,lemmas,results) = 
                      (case evDed(d,env,ab) of
                          v as propVal(P) => getArgVals(rest,P::lemmas,v::results))
               val (arg_vals,lemmas) = getArgVals(args,[],[])
           in
             (case mval of 
                 primMethodVal(m) => applyPrimMethod(m,arg_vals,AB.augment(ab,lemmas))
               | methClosVal({params,body,env=cenv}) => 
                        evDed(body,ref(extend(!cenv,ListPair.zip(params,map bind arg_vals))),
                             AB.augment(ab,lemmas))
               | _ => evError("Non-method value in the leftmost position of a method application."))
           end
      | evDed(A.assumeDed(hyp,body),env,ab) = 
          (case evExp(hyp,env,ab) of
              propVal(P) => (case evDed(body,env,AB.insert(P,ab)) of
                                propVal(Q) => propVal(A.cond(P,Q)))
            | _ => evError("The hypothesis of a hypothetical deduction must be a proposition."))
      | evDed(A.supAbDed(hyp,body),env,ab) = 
          (case evExp(hyp,env,ab) of
              propVal(P) => (case evDed(body,env,AB.insert(P,ab)) of
                                propVal(falseProp) => propVal(A.neg(P))
                              | _ => evError("Suppose-absurd failed to produce a contradiction."))
            | _ => evError("The hypothesis of a suppose-absurd must be a proposition."))
      | evDed(A.letDed([],body),env,ab) = evDed(body,env,ab)
      | evDed(A.letDed((id,A.expression(e))::rest,body),env,ab) = 
          evDed(A.letDed(rest,body),ref(extend(!env,[(id,bind(evExp(e,env,ab)))])),ab)
      | evDed(A.letDed((id,A.deduction(d))::rest,body),env,ab) = 
          (case evDed(d,env,ab) of
              propVal(P) => evDed(A.letDed(rest,body),ref(extend(!env,[(id,bind(evDed(d,env,ab)))])),
                                  AB.insert(P,ab)))
      | evDed(A.byDed(e,d),env,ab) = 
          (case (evDed(d,env,ab),evExp(e,env,ab)) of
              (v as propVal(P),propVal(Q)) => if P = Q then v else evError("Failed BY deduction."))
      | evDed(A.seqDed([]),_,_) = evError("Empty sequence of deductions.")
      | evDed(A.seqDed([d]),env,ab) = evDed(d,env,ab)
      | evDed(A.seqDed(d::rest),env,ab) = 
          (case evDed(d,env,ab) of 
              propVal(P) => evDed(A.seqDed(rest),env,AB.insert(P,ab)))
     | evDed(A.matchDed({discriminant,cases}),env,ab) = 
          (case evExp(discriminant,env,ab) of 
               propVal(P) => 
                  let val (env',e) = findMatch(P,map (fn (pat,d) => (pat,A.deduction(d))) cases)
                  in
                    evPhrase(e,ref(augmentEnv(!env,env')),ab)
                  end
            | _ => evError("Attempt to pattern-match a value that is not a proposition."))
   and  
       evPhrase(A.expression(e),env,ab) = evExp(e,env,ab)
     | evPhrase(A.deduction(d),env,ab) = evDed(d,env,ab);


(*****************************************************************************************)

(********      INSTRUMENTED INTERPETER FOR GENERATING CERTIFICATES              **********)

(*****************************************************************************************)

fun evDedAndExpand(A.methodAppDed(E,args),env,ab) = 
      let val (M,_) = evExpAndExpand(E,env,ab)
          val arg_vals_and_ded_options = map (fn phr => evPhraseAndExpand(phr,env,ab)) args
          val arg_vals = map #1 arg_vals_and_ded_options 
          fun getLemmasAndProofs([],lemmas,proofs) = (lemmas,proofs)
            | getLemmasAndProofs((propVal(P),SOME(D))::rest,lemmas,proofs) = 
                getLemmasAndProofs(rest,P::lemmas,D::proofs)
            | getLemmasAndProofs(_::rest,lemmas,proofs) = getLemmasAndProofs(rest,lemmas,proofs)
          val (lemmas,proofs) = getLemmasAndProofs(arg_vals_and_ded_options,[],[])
      in
         (case M of
             primMethodVal(m) => 
                (applyPrimMethod(m,map #1 arg_vals_and_ded_options,AB.augment(ab,lemmas)),
                 SOME(A.seqDed(rev(A.methodAppDed(A.constExp(A.primMethodConst(m)),
                              (map (fn propVal(P) => A.expression(A.propExp(P))) arg_vals))::proofs))))
           | methClosVal({params,body,env=cenv}) => 
               let val (res,SOME(proof'))= evDedAndExpand(body,ref(extend(!cenv,
                                       ListPair.zip(params,map bind arg_vals))),AB.augment(ab,lemmas))
               in 
                   (res,SOME(A.seqDed(rev(proof'::proofs))))
               end)
      end 
  | evDedAndExpand(A.assumeDed(hyp,body),env,ab) = 
      (case evExp(hyp,env,ab) of
          propVal(P) => 
            (case evDedAndExpand(body,env,AB.insert(P,ab)) of
               (propVal(Q),SOME proof) => (propVal(A.cond(P,Q)),
                                           SOME(A.seqDed[A.assumeDed(A.propExp(P),proof)]))
        | _ => evError("The hypothesis of a hypothetical deduction must be a proposition.")))
  | evDedAndExpand(A.supAbDed(hyp,body),env,ab) = 
          (case evExp(hyp,env,ab) of
              propVal(P) => (case evDedAndExpand(body,env,AB.insert(P,ab)) of
                                (propVal(falseProp),SOME proof) => 
                                  (propVal(A.neg(P)),SOME(A.seqDed[A.supAbDed(A.propExp(P),proof)]))
                              | _ => evError("Suppose-absurd failed to produce a contradiction."))
            | _ => evError("The hypothesis of a suppose-absurd must be a proposition."))
  | evDedAndExpand(A.letDed(bindings,body),env,ab) = 
     let fun getEnvABAndProofs([],env,ab,proofs) = (env,ab,proofs)
           | getEnvABAndProofs((id,A.expression(e))::rest,env,ab,proofs) = 
                getEnvABAndProofs(rest,ref(extend(!env,[(id,bind(evExp(e,env,ab)))])),ab,proofs)
           | getEnvABAndProofs((id,A.deduction(d))::rest,env,ab',proofs) = 
                let val (v as propVal(P),SOME proof) = evDedAndExpand(d,env,ab') 
                in
                  getEnvABAndProofs(rest,ref(extend(!env,[(id,bind(v))])),AB.insert(P,ab'),proof::proofs)
                end
         val (env',ab',proofs) = getEnvABAndProofs(bindings,env,ab,[])
         val (res,SOME(A.seqDed(proofs'))) = evDedAndExpand(body,env',ab')
     in   
       (res,SOME(A.seqDed(rev(proofs)@proofs')))
     end            
  | evDedAndExpand(A.byDed(e,d),env,ab) = 
      (case (evDedAndExpand(d,env,ab),evExp(e,env,ab)) of
        ((v as propVal(P),p_opt),propVal(Q)) => if P = Q then (v,p_opt) else evError("Failed BY deduction."))
  | evDedAndExpand(A.seqDed(deds),env,ab) =
     let fun f([D],proofs,ab) = (case evDedAndExpand(D,env,ab) of
                                   (v,SOME(proof')) => 
                                       (v,SOME(A.seqDed(rev(proof'::proofs)))))
           | f(D::rest,proofs,ab) = 
                 (case evDedAndExpand(D,env,ab) of
                    (propVal(P),SOME(proof')) => f(rest,proof'::proofs,AB.insert(P,ab)))
     in
        f(deds,[],ab)
     end
  | evDedAndExpand(A.matchDed({discriminant,cases}),env,ab) = 
      (case evExp(discriminant,env,ab) of 
               propVal(P) => let val (env',phr) = findMatch(P,map (fn (pat,d) => (pat,A.deduction(d))) cases)
                             in
                               evPhraseAndExpand(phr,ref(augmentEnv(!env,env')),ab)
                             end
            | _ => evError("Attempt to pattern-match a value that is not a proposition."))
and evExpAndExpand(e,env,ab) = (evExp(e,env,ab),NONE)
and evPhraseAndExpand(A.expression(e),env,ab) = evExpAndExpand(e,env,ab)
  | evPhraseAndExpand(A.deduction(d),env,ab) = evDedAndExpand(d,env,ab)


(****************************************************************************************)

(******  INSTRUMENTED INTERPETER FOR COMPUTING THE EVALUATION COST OF A PHRASE     ******)

(****************************************************************************************)

fun primFunCost(A.eqFun) = 1 
  | primFunCost(_) = 0;

fun vcDed(A.methodAppDed(mexp,args),env,ab) = 
      let val (mval,mcost) = vcExp(mexp,env,ab)
          fun argValsCostAndLemmas([],vals,cost,lemmas) = (rev(vals),cost,lemmas)
            | argValsCostAndLemmas(A.expression(e)::rest,vals,cost,lemmas) =  
                 let val (eval,ecost) = vcExp(e,env,ab)
                 in
                    argValsCostAndLemmas(rest,eval::vals,cost+ecost,lemmas)
                 end
            | argValsCostAndLemmas(A.deduction(d)::rest,vals,cost,lemmas) = 
                (case vcDed(d,env,ab) of
                    (v as propVal(P),dcost) => argValsCostAndLemmas(rest,v::vals,cost+dcost,P::lemmas))
          val (arg_vals,arg_cost,lemmas) = argValsCostAndLemmas(args,[],0,[])
      in
         (case mval of
             primMethodVal(m) => (applyPrimMethod(m,arg_vals,AB.augment(ab,lemmas)),1+mcost+arg_cost)
           | methClosVal({params,body,env=cenv}) => 
               let val (bval,bcost) = vcDed(body,ref(extend(!cenv,ListPair.zip(params,map bind arg_vals))),
                                           AB.augment(ab,lemmas))
               in
                  (bval,mcost+arg_cost+bcost+1)
               end)
      end
  | vcDed(A.assumeDed(hyp,body),env,ab) = 
      (case vcExp(hyp,env,ab) of
          (propVal(P),hyp_cost) => (case vcDed(body,env,AB.insert(P,ab)) of
                                       (propVal(Q),bcost) => (propVal(A.cond(P,Q)),hyp_cost+bcost+1))
        | _ => evError("The hypothesis of a hypothetical deduction must be a proposition."))
  | vcDed(A.supAbDed(hyp,body),env,ab) = 
      (case vcExp(hyp,env,ab) of
          (propVal(P),hyp_cost) => (case vcDed(body,env,AB.insert(P,ab)) of
                                       (propVal(falseProp),bcost) => (propVal(A.neg(P)),bcost+hyp_cost+1))
        | _ => evError("The hypothesis of a suppose-absurd must be a proposition."))
  | vcDed(A.letDed(bindings,body),env,ab) = 
      let fun getEnvABAndCost([],env,ab,cost) = (env,ab,cost)
            | getEnvABAndCost((id,A.expression(e))::rest,env,ab,cost) = 
                 let val (v,c) = vcExp(e,env,ab)
                 in
                    getEnvABAndCost(rest,ref(extend(!env,[(id,bind(v))])),ab,cost+c)
                 end
            | getEnvABAndCost((id,A.deduction(d))::rest,env,ab',cost) = 
                 let val (v as propVal(P),c) = vcDed(d,env,ab')
                 in
                   getEnvABAndCost(rest,ref(extend(!env,[(id,bind(v))])),AB.insert(P,ab'),c+cost)
                 end
          val (env',ab',bindings_cost) = getEnvABAndCost(bindings,env,ab,0)
          val (res,body_cost) = vcDed(body,env',ab')
      in
         (res,bindings_cost + body_cost + 1)
      end
  | vcDed(A.byDed(e,d),env,ab) = 
      (case (vcDed(d,env,ab),vcExp(e,env,ab)) of
          ((v as propVal(P),d_cost),(propVal(Q),e_cost)) => if P = Q then (v,d_cost+e_cost+1) else
                                                            evError("Failed BY deduction."))
  | vcDed(A.seqDed(deds),env,ab) = 
      let fun f([D],ab,cost) = (case vcDed(D,env,ab) of
                                   (v,c) => (v,c+cost))
            | f(D::rest,ab,cost) = (case vcDed(D,env,ab) of
                                       (propVal(P),c) => f(rest,AB.insert(P,ab),c+cost))
      in
         f(deds,ab,0)
      end
  | vcDed(A.matchDed({discriminant,cases}),env,ab) = 
      (case vcExp(discriminant,env,ab) of 
          (propVal(P),discr_cost) => let val (env',phr,match_cost) = 
                                                  findMatchAndCost(P,
                                                     map (fn (pat,d) => (pat,A.deduction(d))) cases)
                                         val (res,case_cost) = vcPhrase(phr,ref(augmentEnv(!env,env')),ab)
                                     in
                                        (res,discr_cost+match_cost+case_cost+1)
                                     end
        | _ => evError("Attempt to pattern-match a value that is not a proposition."))
and
    vcExp(A.constExp(A.primMethodConst(m)),_,_) = (primMethodVal(m),0)
  | vcExp(A.constExp(A.primFunConst(f)),_,_) = (primFunVal(f),0)
  | vcExp(A.constExp(A.trueConst),_,_) = (propVal(A.trueProp),0)
  | vcExp(A.constExp(A.falseConst),_,_) = (propVal(A.falseProp),0)
  | vcExp(A.constExp(A.propAtomConst(A)),_,_) = (propVal(A.atom(A)),0)
  | vcExp(A.idExp(id),env,_) = (case (!env)(id) of 
                                       unBound => evError("Unbound identifier: "^id)
                                     | bind(v) => (v,1))
  | vcExp(A.funExp({params,body}),cenv,_) = (funClosVal({params=params,body=body,env=cenv}),1)
  | vcExp(A.methodExp({params,body}),cenv,_) = (methClosVal({params=params,body=body,env=cenv}),1)
  | vcExp(A.funAppExp(fexp,args),env,ab) = 
     let val ((fval,fcost),arg_vals_and_costs) = (vcExp(fexp,env,ab),map (fn p => vcPhrase(p,env,ab)) args)
         val (arg_vals,arg_costs) = ListPair.unzip(arg_vals_and_costs)
         val arg_cost = List.foldl op+ 0 arg_costs
     in      
       (case fval of
           primFunVal(f) => (applyPrimFun(f,arg_vals,ab),(primFunCost(f))+fcost+arg_cost)
         | funClosVal({params,body,env=cenv}) => 
            let val (bval,bcost) = vcExp(body,ref(extend(!cenv,ListPair.zip(params,map bind arg_vals))),ab)
            in
               (bval,bcost+fcost+arg_cost+1)
            end
         | _ => evError("Non-function value in the leftmost position of a function application."))
     end
  | vcExp(A.letExp(bindings,body),env,ab) = 
      let fun getEnvAndCost([],env,cost) = (env,cost)
            | getEnvAndCost((id,phr)::rest,env,cost) = 
               let val (v,c) = vcPhrase(phr,env,ab)
               in
                  getEnvAndCost(rest,ref(extend(!env,[(id,bind(v))])),c+cost)
               end
          val (env',bindings_cost) = getEnvAndCost(bindings,env,0)
          val (body_val,body_cost) = vcExp(body,env',ab)
      in
         (body_val,body_cost+bindings_cost+1)
      end
  | vcExp(A.fixExp(x,body),env,ab) =
          let val rec_env = ref(!env)
              val (rec_value,cost) = vcExp(body,rec_env,ab)
              val new_env = extend(!env,[(x,bind rec_value)])
              val _ = rec_env := new_env
          in
             (rec_value,cost+1)
          end
  | vcExp(A.matchExp({discriminant,cases}),env,ab) = 
      (case vcExp(discriminant,env,ab) of 
          (propVal(P),discr_cost) => let val (env',phr,match_cost) = 
                                                  findMatchAndCost(P,
                                                     map (fn (pat,e) => (pat,A.expression(e))) cases)
                                         val (res,case_cost) = vcPhrase(phr,ref(augmentEnv(!env,env')),ab)
                                     in
                                        (res,discr_cost+match_cost+case_cost+1)
                                     end
        | _ => evError("Attempt to pattern-match a value that is not a proposition."))
  | vcExp(A.seqExp(exps),env,ab) = 
      let fun f([e],c) = let val (v,c') = vcExp(e,env,ab)
                         in
                            (v,c+c')
                         end
            | f(e::rest,c) = let val (_,c') = vcExp(e,env,ab)  
                             in
                                f(rest,c+c')
                             end
      in
         f(exps,0)
      end
and
    vcPhrase(A.expression(e),env,ab) = vcExp(e,env,ab)
  | vcPhrase(A.deduction(d),env,ab) = vcDed(d,env,ab);


(*********************************************************************************************)

(******  INSTRUMENTED INTERPETER FOR CALCULATING THE COMPUTATIONAL COST OF A PHRASE     ******)

(*********************************************************************************************)

fun vc_E_Ded(A.methodAppDed(mexp,args),env,ab) = 
      let val (mval,mcost) = vc_E_Exp(mexp,env,ab)
          fun argValsCostAndLemmas([],vals,cost,lemmas) = (rev(vals),cost,lemmas)
            | argValsCostAndLemmas(A.expression(e)::rest,vals,cost,lemmas) =  
                 let val (eval,ecost) = vc_E_Exp(e,env,ab)
                 in
                    argValsCostAndLemmas(rest,eval::vals,cost+ecost,lemmas)
                 end
            | argValsCostAndLemmas(A.deduction(d)::rest,vals,cost,lemmas) = 
                (case vc_E_Ded(d,env,ab) of
                    (v as propVal(P),dcost) => argValsCostAndLemmas(rest,v::vals,cost+dcost,P::lemmas))
          val (arg_vals,arg_cost,lemmas) = argValsCostAndLemmas(args,[],0,[])
      in
         (case mval of
             primMethodVal(m) => (applyPrimMethod(m,arg_vals,AB.augment(ab,lemmas)),mcost+arg_cost)
           | methClosVal({params,body,env=cenv}) => 
               let val (bval,bcost) = vc_E_Ded(body,ref(extend(!cenv,ListPair.zip(params,map bind arg_vals))),
                                           AB.augment(ab,lemmas))
               in
                  (bval,mcost+arg_cost+bcost+1)
               end)
      end
  | vc_E_Ded(A.assumeDed(hyp,body),env,ab) = 
      (case vc_E_Exp(hyp,env,ab) of
          (propVal(P),hyp_cost) => (case vc_E_Ded(body,env,AB.insert(P,ab)) of
                                       (propVal(Q),bcost) => (propVal(A.cond(P,Q)),hyp_cost+bcost))
        | _ => evError("The hypothesis of a hypothetical deduction must be a proposition."))
  | vc_E_Ded(A.supAbDed(hyp,body),env,ab) = 
      (case vc_E_Exp(hyp,env,ab) of
          (propVal(P),hyp_cost) => (case vc_E_Ded(body,env,AB.insert(P,ab)) of
                                       (propVal(falseProp),bcost) => (propVal(A.neg(P)),bcost+hyp_cost))
        | _ => evError("The hypothesis of a suppose-absurd must be a proposition."))
  | vc_E_Ded(A.letDed(bindings,body),env,ab) = 
      let fun getEnvABAndCost([],env,ab,cost) = (env,ab,cost)
            | getEnvABAndCost((id,A.expression(e))::rest,env,ab,cost) = 
                 let val (v,c) = vc_E_Exp(e,env,ab)
                 in
                    getEnvABAndCost(rest,ref(extend(!env,[(id,bind(v))])),ab,cost+c)
                 end
            | getEnvABAndCost((id,A.deduction(d))::rest,env,ab',cost) = 
                 let val (v as propVal(P),c) = vc_E_Ded(d,env,ab')
                 in
                   getEnvABAndCost(rest,ref(extend(!env,[(id,bind(v))])),AB.insert(P,ab'),c+cost)
                 end
          val (env',ab',bindings_cost) = getEnvABAndCost(bindings,env,ab,0)
          val (res,body_cost) = vc_E_Ded(body,env',ab')
      in
         (res,bindings_cost + body_cost + 1)
      end
  | vc_E_Ded(A.byDed(e,d),env,ab) = 
      (case (vc_E_Ded(d,env,ab),vc_E_Exp(e,env,ab)) of
          ((v as propVal(P),d_cost),(propVal(Q),e_cost)) => if P = Q then (v,d_cost+e_cost+1) else
                                                            evError("Failed BY deduction."))
  | vc_E_Ded(A.seqDed(deds),env,ab) = 
      let fun f([D],ab,cost) = (case vc_E_Ded(D,env,ab) of
                                   (v,c) => (v,c+cost))
            | f(D::rest,ab,cost) = (case vc_E_Ded(D,env,ab) of
                                       (propVal(P),c) => f(rest,AB.insert(P,ab),c+cost))
      in
         f(deds,ab,0)
      end
  | vc_E_Ded(A.matchDed({discriminant,cases}),env,ab) = 
      (case vc_E_Exp(discriminant,env,ab) of 
          (propVal(P),discr_cost) => let val (env',phr,match_cost) = 
                                                  findMatchAndCost(P,
                                                     map (fn (pat,d) => (pat,A.deduction(d))) cases)
                                         val (res,case_cost) = vc_E_Phrase(phr,ref(augmentEnv(!env,env')),ab)
                                     in
                                        (res,discr_cost+match_cost+case_cost+1)
                                     end
        | _ => evError("Attempt to pattern-match a value that is not a proposition."))
and
    vc_E_Exp(e,env,ab) = vcExp(e,env,ab)
and
    vc_E_Phrase(A.expression(e),env,ab) = vcExp(e,env,ab)
  | vc_E_Phrase(A.deduction(d),env,ab) = vc_E_Ded(d,env,ab);


(*****************************************************************************************)

(******  INSTRUMENTED INTERPETER FOR CALCULATING THE DEDUCTIVE COST OF A PHRASE     ******)

(*****************************************************************************************)

fun vc_D_Ded(A.methodAppDed(mexp,args),env,ab) = 
      let val (mval,mcost) = vc_D_Exp(mexp,env,ab)
          fun argValsCostAndLemmas([],vals,cost,lemmas) = (rev(vals),cost,lemmas)
            | argValsCostAndLemmas(A.expression(e)::rest,vals,cost,lemmas) =  
                 let val (eval,ecost) = vc_D_Exp(e,env,ab)
                 in
                    argValsCostAndLemmas(rest,eval::vals,cost+ecost,lemmas)
                 end
            | argValsCostAndLemmas(A.deduction(d)::rest,vals,cost,lemmas) = 
                (case vc_D_Ded(d,env,ab) of
                    (v as propVal(P),dcost) => argValsCostAndLemmas(rest,v::vals,cost+dcost,P::lemmas))
          val (arg_vals,arg_cost,lemmas) = argValsCostAndLemmas(args,[],0,[])
      in
         (case mval of
             primMethodVal(m) => (applyPrimMethod(m,arg_vals,AB.augment(ab,lemmas)),1+mcost+arg_cost)
           | methClosVal({params,body,env=cenv}) => 
               let val (bval,bcost) = vc_D_Ded(body,ref(extend(!cenv,ListPair.zip(params,map bind arg_vals))),
                                           AB.augment(ab,lemmas))
               in
                  (bval,mcost+arg_cost+bcost)
               end)
      end
  | vc_D_Ded(A.assumeDed(hyp,body),env,ab) = 
      (case vc_D_Exp(hyp,env,ab) of
          (propVal(P),hyp_cost) => (case vc_D_Ded(body,env,AB.insert(P,ab)) of
                                       (propVal(Q),bcost) => (propVal(A.cond(P,Q)),1+hyp_cost+bcost))
        | _ => evError("The hypothesis of a hypothetical deduction must be a proposition."))
  | vc_D_Ded(A.supAbDed(hyp,body),env,ab) = 
      (case vc_D_Exp(hyp,env,ab) of
          (propVal(P),hyp_cost) => (case vc_D_Ded(body,env,AB.insert(P,ab)) of
                                       (propVal(falseProp),bcost) => (propVal(A.neg(P)),1+bcost+hyp_cost))
        | _ => evError("The hypothesis of a suppose-absurd must be a proposition."))
  | vc_D_Ded(A.letDed(bindings,body),env,ab) = 
      let fun getEnvABAndCost([],env,ab,cost) = (env,ab,cost)
            | getEnvABAndCost((id,A.expression(e))::rest,env,ab,cost) = 
                 let val (v,c) = vc_D_Exp(e,env,ab)
                 in
                    getEnvABAndCost(rest,ref(extend(!env,[(id,bind(v))])),ab,cost+c)
                 end
            | getEnvABAndCost((id,A.deduction(d))::rest,env,ab',cost) = 
                 let val (v as propVal(P),c) = vc_D_Ded(d,env,ab')
                 in
                   getEnvABAndCost(rest,ref(extend(!env,[(id,bind(v))])),AB.insert(P,ab'),c+cost)
                 end
          val (env',ab',bindings_cost) = getEnvABAndCost(bindings,env,ab,0)
          val (res,body_cost) = vc_D_Ded(body,env',ab')
      in
         (res,bindings_cost + body_cost)
      end
  | vc_D_Ded(A.byDed(e,d),env,ab) = 
      (case (vc_D_Ded(d,env,ab),vc_D_Exp(e,env,ab)) of
          ((v as propVal(P),d_cost),(propVal(Q),e_cost)) => if P = Q then (v,d_cost+e_cost) else
                                                            evError("Failed BY deduction."))
  | vc_D_Ded(A.seqDed(deds),env,ab) = 
      let fun f([D],ab,cost) = (case vc_D_Ded(D,env,ab) of
                                   (v,c) => (v,c+cost))
            | f(D::rest,ab,cost) = (case vc_D_Ded(D,env,ab) of
                                       (propVal(P),c) => f(rest,AB.insert(P,ab),c+cost))
      in
         f(deds,ab,0)
      end
  | vc_D_Ded(A.matchDed({discriminant,cases}),env,ab) = 
      (case vc_D_Exp(discriminant,env,ab) of 
          (propVal(P),discr_cost) => let val (env',phr,match_cost) = 
                                                  findMatchAndCost(P,
                                                     map (fn (pat,d) => (pat,A.deduction(d))) cases)
                                         val (res,case_cost) = vc_D_Phrase(phr,ref(augmentEnv(!env,env')),ab)
                                     in
                                        (res,case_cost)
                                     end
        | _ => evError("Attempt to pattern-match a value that is not a proposition."))
and
    vc_D_Exp(e,env,ab) = (evExp(e,env,ab),0)
and
    vc_D_Phrase(A.expression(e),env,ab) = vc_D_Exp(e,env,ab)
  | vc_D_Phrase(A.deduction(d),env,ab) = vc_D_Ded(d,env,ab);

end;  (* end of Semant structure *)
